(function() {


    /////https://chriscavs.github.io/bounds-demo///////////
    ////library slightly changed ... usecase intersectiobObserver ///////////
    const checkForObserver = () => {
        if (!("IntersectionObserver" in window)) {
            throw new Error(`
      bounds.js requires IntersectionObserver on the global object.
      IntersectionObserver is unavailable in IE and other older
      versions of browsers.
      See https://github.com/ChrisCavs/bounds.js/blob/master/README.md
      for more compatibility information.
    `);
            return false;
        }
        return true;
    };

    const getMargins = (margins = {}) => {
        const { top = 0, right = 0, bottom = 0, left = 0 } = margins;
        return `${top}px ${right}px ${bottom}px ${left}px`;
    };

    const noOp = () => {};

    const Bound = options => {
        return new Boundary(options);
    };

    class Boundary {
        constructor({ root, margins, threshold, onEmit } = {}) {
                //checkForObserver();
                const marginString = getMargins(margins);
                const options = {
                    root: root || null,
                    rootMargin: marginString,
                    threshold: threshold || 0.0
                };
                this.nodes = [];
                this.onEmit = onEmit || noOp;
                this.observer = new IntersectionObserver(this._emit.bind(this), options);
            } // API

        watch(el, onEnter = noOp, onLeave = noOp) {
            const data = {
                el,
                onEnter,
                onLeave
            };
            this.nodes.push(data);
            this.observer.observe(el);
            return data;
        }

        unWatch(el) {
            const index = this._findByNode(el, true);

            if (index !== -1) {
                this.nodes.splice(index, 1);
                this.observer.unobserve(el);
            }

            return this;
        }

        check(el) {
            const data = this._findByNode(el) || {};
            return data.history;
        }

        clear() {
            this.nodes = [];
            this.observer.disconnect();
            return this;
        }

        static checkCompatibility() {
                checkForObserver();
            } // HELPERS

        _emit(events) {
            const actions = events.map(event => {
                const data = this._findByNode(event.target);

                const ratio = event.intersectionRatio;
                data.history = event.isIntersecting;
                event.isIntersecting ? data.onEnter(ratio) : data.onLeave(ratio);
                return {
                    dropTargetEl: document.getElementById(this.dropTargetId),
                    el: event.target,
                    inside: event.isIntersecting,
                    outside: !event.isIntersecting,
                    ratio: event.intersectionRatio
                };
            });
            this.onEmit(actions);
        }

        _findByNode(el, returnIndex = false) {
            const func = returnIndex ? "findIndex" : "find";
            return this.nodes[func](node => {
                return node.el.isEqualNode(el);
            });
        }
    } /////end of bounds///////////////////////////////////
    //////unloading event////////////////////////////////

    function unloadDragDrop(hypeDocument, element, event) {
        console.log(this.callback.boundsInstances)
        Object.values(this.callback.boundsInstances).forEach(
            boundsInstance => {
                boundsInstance.clear();
            }
        );
        this.callback.boundsInstances = {};
        console.log(this.callback.boundsInstances)

        window.removeEventListener(
            "resize",
            this.callback.initIntersectInterSectionObservers
        );

    } ///////end unloading//////////////////////////////////////

    function dragDrop(hypeDocument, element, event) {


        function initDropTargetMutationObserver(dropTarget) {
            hypeDocument.startMutationObserver(
                dropTarget,
                function(mutation) {
                    Object.keys(boundsInstances).forEach(observerInstanceKey => {
                        if (
                            boundsInstances[observerInstanceKey]["dropTargetId"] ===
                            dropTarget.id
                        ) {
                            const dragTarget = hypeDocument.getElementById(
                                boundsInstances[observerInstanceKey]["dragTargetId"]
                            );
                            boundsInstances[observerInstanceKey].clear();
                            setInterSectionObserver([dragTarget, [dropTarget]]);
                        }
                    });
                }, {
                    id: dropTarget.id,
                    updaterate: 2,
                    attributes: true,
                    attributeFilter: ["style"]
                }
            );
        } /////

        const initInterSectionObservers = (unloadDragDrop.initIntersectInterSectionObservers = function() {
            setTimeout(function() {
                dragnDropPairs.forEach(function(dragDropPair) {
                    setInterSectionObserver(dragDropPair);
                });
            }, 1);
        }); /////

        function filldragnDropPairs(dragEls) {
            dragEls.forEach(function(dragEl) {


                const _dropClasses = dragEl.dataset.intersecttarget.split(" | ");

                const dropEls = getDropEls(_dropClasses);
                dragnDropPairs.push([dragEl, dropEls]);
            });
        } /////

        function getDropTargetMargins(dropTarget, SceneEl) {
            const margins = {};
            const scrollLeft =
                window.pageXOffset || document.documentElement.scrollLeft;
            const scrollTop = window.pageYOffset || document.documentElement.scrollTop;
            const dropTargetViewportCoords = dropTarget.getBoundingClientRect();
            const dropTargetTop = dropTargetViewportCoords.top + scrollTop;
            const dropTargetLeft = dropTargetViewportCoords.left + scrollLeft;
            const dropTargetRight = dropTargetViewportCoords.right + scrollLeft;
            const sceneViewportCoords = SceneEl.getBoundingClientRect();
            const sceneTop = sceneViewportCoords.top + scrollTop;
            const sceneLeft = sceneViewportCoords.left + scrollLeft;
            const sceneRight = sceneViewportCoords.right + scrollLeft;
            margins.left = Math.abs(dropTargetLeft - sceneLeft) * -1;
            margins.top = Math.abs(dropTargetTop - sceneTop) * -1;
            margins.bottom =
                Math.abs(
                    sceneViewportCoords.height -
                    dropTargetViewportCoords.height -
                    dropTargetTop
                ) * -1;
            margins.right = Math.abs(dropTargetRight - sceneRight) * -1;
            return margins;
        } /////

        function setInterSectionObserver(dragDropPair) {
            const dragTarget = dragDropPair[0];
            const dropTargets = dragDropPair[1];

            if (dropTargets.length === 0) {
                console.log('No droptargets for ' + dragTarget.id)
                return
            };

            dropTargets.forEach(function(dropTarget) {
                const observerInstanceKey = dropTarget.id + " | " + dragTarget.id;

                if (boundsInstances[observerInstanceKey]) {
                    boundsInstances[observerInstanceKey].clear();
                }

                const margins = getDropTargetMargins(dropTarget, currSceneEl);
                boundsInstances[observerInstanceKey] = Bound({
                    root: currSceneEl,
                    margins: margins,
                    threshold: 0,
                    onEmit: actions => {
                        //console.log(actions)
                        if (actions.some(action => action.inside)) {
                            const currIntersection = actions[0];
                            const custombehaviorNameToTrigger = currIntersection.dropTargetEl.dataset.intersectcustombehavior;
                            if (custombehaviorNameToTrigger) {
                                console.log("source intersection, custombehaviorNameToTrigger: " + custombehaviorNameToTrigger);

                                hypeDocument.triggerCustomBehaviorNamed(
                                    custombehaviorNameToTrigger
                                );


                            } else {
                                console.log("source intersection, custombehaviorNameToTrigger: " + custombehaviorNameToTrigger);
                            }
                        };
                        if (actions.some(action => action.outside)) {

                            const currIntersection = actions[0];
                            const custombehaviorNameToTrigger = currIntersection.dropTargetEl.dataset.intersectoffcustombehavior;
                            if (custombehaviorNameToTrigger) {
                                console.log("source intersectionoff, custombehaviorNameToTrigger: " + custombehaviorNameToTrigger);

                                hypeDocument.triggerCustomBehaviorNamed(
                                    custombehaviorNameToTrigger
                                );


                            } else {
                                console.log("source intersectionoff, custombehaviorNameToTrigger: " + custombehaviorNameToTrigger);

                            }
                        }

                    }
                });
                boundsInstances[observerInstanceKey].watch(dragTarget);
                boundsInstances[observerInstanceKey]["dropTargetId"] = dropTarget.id;
                boundsInstances[observerInstanceKey]["dragTargetId"] = dragTarget.id;
                initDropTargetMutationObserver(dropTarget);
            });
        } /////

        function getDragEls() {
            const dragEls = [...currSceneEl.querySelectorAll("[data-intersect]")];
            return dragEls;
        } /////

        function getDropEls(ArrayOfClasses) {
            let tmpArr = [];
            ArrayOfClasses.forEach(function(cssClass) {
                tmpArr.push([...currSceneEl.querySelectorAll("." + cssClass)]);
            });
            return [].concat(...tmpArr);
        } /////


        //global
        const currSceneEl = document.getElementById(hypeDocument.currentSceneId());
        const dragnDropPairs = []; //[[targetE, [dropel, dropel, ...]], ..]

        const boundsInstances = (unloadDragDrop.boundsInstances = {});


        //////////////////////////main thread///////////

        const dragEls = getDragEls(); //
        if (!checkForObserver()) { return };


        filldragnDropPairs(dragEls);

        initInterSectionObservers(); //


        window.addEventListener("resize", initInterSectionObservers);
    }

    if ("HYPE_eventListeners" in window === false) {
        window.HYPE_eventListeners = Array();
    }

    window.HYPE_eventListeners.push({
        type: "HypeSceneLoad",
        callback: dragDrop
    });
    window.HYPE_eventListeners.push({
        type: "HypeSceneUnload",
        callback: unloadDragDrop
    });
})()